#include "../include/nicky_bdo.h"

int isNickyBDOFile(char* fileName)
{
    if ((fileName[0] < '0') || (fileName[0] > '9'))
    {
        return 0;
    }
    else
    {
        int lastIDX  = strlen(fileName) - 1;
        if (fileName[lastIDX-2] != 'd' || fileName[lastIDX-1] != 'd' || fileName[lastIDX] != 's')
        {
            return 0;
        }
    }
    return 1;
}
char* getNickyBDORealName(char* fileToPatchName)
{
    int i = 0;
    while(renames[i][0] && renames[i][1])
    {
        if (strcmp(renames[i][0],fileToPatchName) == 0)
        {
            printf(" -> %s ",renames[i][1]);
            return renames[i][1];
        }
        i++;
    }
    return NULL;
}
int isNickyBDOSkipableFile(FileBlock* fileToPatch)
{
    if (strcmpi(fileToPatch->fileName,"partcutdesc.xml") == 0)
    {
        char* pathToPartcutdesc = (char*) malloc(strlen("files_to_patch\\") + strlen(fileToPatch->originalPath) + strlen(fileToPatch->fileName) + 1);

        strcpy(pathToPartcutdesc,"files_to_patch\\");
        strcat(pathToPartcutdesc,fileToPatch->originalPath);
        strcat(pathToPartcutdesc,fileToPatch->fileName);

        if (getFileSizeByName(pathToPartcutdesc) == NICKY_BDO_PARTCUTDESC_FILE_SIZE)
        {
            free(pathToPartcutdesc);
            return 1;
        }
        free(pathToPartcutdesc);
    }
    if (!isNickyBDOFile(fileToPatch->fileName))
    {
        return 0;
    }
    int i = 0;
    while(skipableFiles[i])
    {
        if (strcmp(skipableFiles[i],fileToPatch->fileName) == 0)
        {
            return 1;
        }
        i++;
    }
    return 0;
}


char** readCommaSeparatedFile(char* pathToFile, int* totalEntriesReturn)
{
    FILE* commaSeparatedFile = fopen(pathToFile,"rb");

    char* fileAsStr = NULL;

    int i = 0;
    int totalEntries = 0;

    if (commaSeparatedFile)
    {
        fileAsStr = (char*) malloc(getFileSize(commaSeparatedFile) * sizeof(char));

        // Reads the file and counts how many '.' (from .pac) it has, giving the total number of file names present
        while(fread(&fileAsStr[i],sizeof(char),1,commaSeparatedFile))
        {
            if (fileAsStr[i] == '.')
            {
                totalEntries++;
            }
            i++;
        }
    }
    if (totalEntries == 0)
    {
        printf("\nNo entries found in \"%s\".\n", pathToFile);
        return NULL;
    }

    char** entries = (char**) malloc(totalEntries * sizeof(char*));

    i = 0;

    // If there is something to extract from the .txt file
    if (commaSeparatedFile != NULL && fileAsStr != NULL && strlen(fileAsStr) > 0)
    {
        // Gets the first file name between ','
        char* nameBetweenCommas = strtok(fileAsStr,",");

        // While there is still files between ','
        while(nameBetweenCommas)
        {
            // Copies the file name to the array
            entries[i] = (char*) malloc(strlen(nameBetweenCommas) + 1);
            strcpy(entries[i],nameBetweenCommas);

            // Gets the next file name
            nameBetweenCommas = strtok(NULL,",");
            // Increases the current array pos to write
            i++;
        }
    }
    (*totalEntriesReturn) = totalEntries;

    return entries;
}

void modifyPartcutdescNickyBDO(FileBlock* fileBlocks, MetaFileInfo* metaFileInfo)
{
    int i = 0;
    int totalNamesToRemove = 0;

    printf("\nReading \"patcher_resources\\nickybdo_fix_holes_files.txt\"...");
    char** namesFromTxtFile = readCommaSeparatedFile("patcher_resources\\nickybdo_fix_holes_files.txt",&totalNamesToRemove);
    printf("\nFound %d entries.\n", totalNamesToRemove);

    if (totalNamesToRemove == 0 && namesFromTxtFile != NULL)
    {
        printf("\nNo files detected in \"patcher_resources\\nickybdo_fix_holes_files.txt\"");
        PAUSE();
        return;
    }

    // If the file nickybdo_fix_holes_files.txt was read successfully
    if (namesFromTxtFile != NULL)
    {
        // We are going to find the file names from nickybdo_fix_holes_files.txt in the fileBlocks array
        // and we are going to set the .needPatch to 1, indicating that it will be removed in the
        // partcutdesc.xml file later

        // Resets the variable for all file blocks, initially
        for(i = 0; i < metaFileInfo->fileBlocksCount; i++)
        {
            fileBlocks[i].needPatch = 0;
        }

        printf("\nSearching for the files found in \"patcher_resources\\nickybdo_fix_holes_files.txt\"");
        int j = 0;
        for(j = 0; j < totalNamesToRemove; j++) // For each file found in the nickybdo_fix_holes_files.txt file
        {
            FileBlock* fileBlockFound = binarySearchFileBlock(namesFromTxtFile[j],fileBlocks,metaFileInfo->fileBlocksCount);

            if (fileBlockFound != NULL)
            {
                fileBlockFound->needPatch = 1;
            }
            else
            {
                printf("\n%s - ", namesFromTxtFile[j]); printColor("Not Found", RED);
            }
        }
        char* root = bdoRootFolder();
        char* pathToOriginalPartcutDesc = concatenate(root,"character\\partcutdesc.xml");

        // If the partcutdesc.xml file was never generated
        if (!fileExists(pathToOriginalPartcutDesc))
        {
            // Creates a new file, with the file blocks marked as "needPath == 1" removed, merged with the original extracted partcutdesc.xml file, also with those files removed.
            generatePartcutdesc(fileBlocks,metaFileInfo);
        }
        else // If the partcutdesc.xml file was already created (by tools like resorepless)
        {
            // Only modify the file, and remove the files
            modifyPartcutdesc(pathToOriginalPartcutDesc,fileBlocks,metaFileInfo);
        }

        free(pathToOriginalPartcutDesc);
        free(root);
    }

    for (i = 0; i < totalNamesToRemove; i++)
    {
        if ((namesFromTxtFile[i] != NULL) && strlen(namesFromTxtFile[i]) > 0)
        {
            free(namesFromTxtFile[i]);
        }
    }
    if (namesFromTxtFile != NULL)
    {
        free(namesFromTxtFile);
    }
}





char* renames[][2] = {
{"100991186.dds","pew_00_ub_0039.dds"},
{"1029840092.dds","pww_00_sho_0001.dds"},
{"1039262907.dds","pvw_00_cloak_0052_hair.dds"},
{"1043928769.dds","phw_02_ub_0002.dds"},
{"1046038268.dds","pew_10_hand_0010.dds"},
{"1063746730.dds","pvw_00_ub_0002_hair.dds"},
{"1089750302.dds","pew_99_ub_0006_02.dds"},
{"1097733778.dds","pbw_10_ub_0007.dds"},
{"1099590690.dds","pew_00_sho_0045.dds"},
{"1102734382.dds","pew_00_lb_0010_dec.dds"},
{"1125428684.dds","pew_10_ub_0010.dds"},
{"1134791207.dds","pbw_00_lb_0017.dds"},
{"1148761656.dds","phw_01_sho_0003_dm.dds"},
{"1241310190.dds","pew_00_lb_0000.dds"},
{"1246446119.dds","phw_02_sho_0001.dds"},
{"1263548148.dds","pbw_00_ub_0015_hair.dds"},
{"128015358.dds","pvw_00_ub_0000.dds"},
{"1294428038.dds","phw_bw_ub_0017.dds"},
{"1295731765.dds","pew_00_lb_0010.dds"},
{"1317877797.dds","pew_00_foot_0010.dds"},
{"1328410138.dds","phw_ew_ub_0009.dds"},
{"1366122890.dds","pkw_00_uw_0001.dds"},
{"1381661087.dds","pew_99_ub_0004_02.dds"},
{"1385462498.dds","phw_00_lb_0039_hair.dds"},
{"1396488354.dds","pkw_00_uw_0001_dec.dds"},
{"1397177222.dds","pew_00_ub_0038_01_hair.dds"},
{"141608490.dds","pew_00_ub_0032.dds"},
{"1416730309.dds","phw_01_sho_0002.dds"},
{"142839219.dds","pvw_03_lb_0001.dds"},
{"1456422651.dds","pew_00_lb_0032_01_hair.dds"},
{"1461547085.dds","pvw_00_ub_0058_00.dds"},
{"1477459045.dds","pvw_00_ub_0003_dec.dds"},
{"1519238861.dds","pvw_00_ub_0052_01.dds"},
{"1564610069.dds","phw_00_lb_0009_01.dds"},
{"1566608110.dds","pkw_ew_ub_0009.dds"},
{"1583947401.dds","phw_bw_ub_0017_hair.dds"},
{"1657787454.dds","pkw_00_ub_0039.dds"},
{"1676321094.dds","phw_02_ub_0001.dds"},
{"1709492249.dds","pbw_00_uw_0001.dds"},
{"1752553477.dds","phw_01_sho_0001.dds"},
{"1767589150.dds","pww_00_ub_0056.dds"},
{"1859759030.dds","pbw_02_lb_0006.dds"},
{"186991410.dds","phm_00_ub_0007.dds"},
{"1874838806.dds","phw_00_hand_0007_dec.dds"},
{"1884415993.dds","pkw_00_ub_0061.dds"},
{"1887005694.dds","phw_01_ub_0002.dds"},
{"1887088388.dds","pew_10_ub_0009.dds"},
{"190688329.dds","pvw_03_hand_0001.dds"},
{"1944530230.dds","pbw_00_ub_0018.dds"},
{"1945400836.dds","pbw_02_ub_0006.dds"},
{"1952449661.dds","pkw_00_lb_0061_hair.dds"},
{"197338074.dds","pww_00_lb_0056.dds"},
{"1974397284.dds","pvw_00_ub_0001.dds"},
{"1974862593.dds","pew_00_ub_0031_01.dds"},
{"1988395424.dds","pew_00_lb_0030.dds"},
{"2015451955.dds","pvw_ew_foot_0009.dds"},
{"2025121141.dds","pew_00_lb_0033.dds"},
{"2055461086.dds","pww_00_ub_0056_hair.dds"},
{"208727022.dds","pvw_00_cloak_0052.dds"},
{"2105536770.dds","pbw_02_ub_0001_01.dds"},
{"2117706239.dds","pew_00_ub_0031.dds"},
{"2133329755.dds","phw_01_lb_0002.dds"},
{"2146457107.dds","phw_ew_ub_0009_hair.dds"},
{"2166737645.dds","pkw_00_lb_0039.dds"},
{"217137700.dds","pvw_bw_ub_0017_01_hair.dds"},
{"2174063315.dds","pew_00_lb_0014.dds"},
{"2188209681.dds","phw_bw_ub_0017_01.dds"},
{"2243436444.dds","pew_00_ub_0032_01_dec.dds"},
{"2250419253.dds","phw_03_sho_0001.dds"},
{"2261774081.dds","phw_01_lb_0001b.dds"},
{"2277543367.dds","pew_00_lb_0039_hair.dds"},
{"2282443151.dds","pew_00_sho_0031.dds"},
{"229754289.dds","phw_99_ub_0001_02.dds"},
{"2298658208.dds","pww_00_lb_0039_hair.dds"},
{"2348765955.dds","pew_00_sho_0032.dds"},
{"2367387927.dds","pww_00_foot_0056.dds"},
{"2394342354.dds","pvw_00_hand_0052.dds"},
{"2397344132.dds","pew_00_lb_0039.dds"},
{"2407061868.dds","pvw_02_foot_0005.dds"},
{"2408996972.dds","pvw_02_lb_0001.dds"},
{"2416015740.dds","phw_00_uw_0001_03_dm_dec.dds"},
{"2416262912.dds","pew_99_ub_0001_02.dds"},
{"2455192927.dds","pkw_00_ub_0001_01.dds"},
{"2478409208.dds","pew_99_ub_0006_dec.dds"},
{"248264298.dds","phw_00_lb_0009_01_dec.dds"},
{"2506496657.dds","pvw_02_ub_0002.dds"},
{"2511301729.dds","phw_01_ub_0003_dm.dds"},
{"2531410653.dds","phw_00_uw_0031_02_dec.dds"},
{"2555568019.dds","pvw_03_ub_0001.dds"},
{"25815453.dds","pew_00_ub_0038.dds"},
{"2618523828.dds","phw_01_lb_0001a.dds"},
{"2624447466.dds","pew_99_ub_0004_dec.dds"},
{"2625270540.dds","pew_01_ub_0001.dds"},
{"2634132228.dds","phw_01_ub_0002_hair.dds"},
{"263671690.dds","pbw_00_ub_0017_hair.dds"},
{"2683995873.dds","pvw_00_lb_0039.dds"},
{"2697968275.dds","phw_01_ub_0001.dds"},
{"2742255835.dds","phw_00_sho_0007_dec.dds"},
{"2745886854.dds","pew_99_ub_0001_dec.dds"},
{"2780978222.dds","pbw_00_ub_0015.dds"},
{"2786132627.dds","pew_00_hand_0032.dds"},
{"2808428804.dds","pvw_00_hand_0000.dds"},
{"2809859456.dds","pvw_00_ub_0058.dds"},
{"2815753251.dds","phw_99_ub_0001_dec.dds"},
{"2837908909.dds","pew_00_foot_0003.dds"},
{"2929245433.dds","pvw_00_cloak_0058.dds"},
{"2929744351.dds","pbw_10_lb_0007.dds"},
{"2931933947.dds","pew_00_ub_0034_01_dm_hair.dds"},
{"2933232231.dds","pew_01_lb_0001.dds"},
{"2963263657.dds","phw_00_ub_0007.dds"},
{"2979213320.dds","pkw_bw_ub_0015.dds"},
{"2985095720.dds","pvw_00_cloak_0002.dds"},
{"2991309761.dds","pbw_02_lb_0001.dds"},
{"3041640327.dds","phw_02_ub_0005.dds"},
{"3125651417.dds","pvw_00_cloak_0034_hair.dds"},
{"3137110903.dds","pew_00_sho_0031_hair.dds"},
{"3143741981.dds","pew_00_ub_0008.dds"},
{"3145493611.dds","phw_00_ub_0030_hair.dds"},
{"3154731346.dds","pbw_00_foot_0046.dds"},
{"3195384492.dds","pkw_00_lb_0039_hair.dds"},
{"3206200443.dds","phw_02_lb_0005.dds"},
{"3285174454.dds","pvw_00_sho_0052.dds"},
{"3294327452.dds","pew_00_cloak_0033.dds"},
{"3314154311.dds","phw_00_lb_0007_dec.dds"},
{"3352794608.dds","phw_00_uw_0001_03_dec.dds"},
{"3360560650.dds","pvw_03_foot_0001.dds"},
{"3368746478.dds","pkw_00_lb_0061.dds"},
{"337553159.dds","pew_01_sho_0001.dds"},
{"3394329051.dds","pkw_bw_ub_0015_hair.dds"},
{"3403189117.dds","pvw_02_ub_0001.dds"},
{"3438578968.dds","pbw_00_lb_0017_dm.dds"},
{"3443865558.dds","pvw_00_sho_0003_01_hair.dds"},
{"3451868287.dds","phw_00_ub_0009_01.dds"},
{"3462953894.dds","pvw_02_ub_0005.dds"},
{"3492574477.dds","pew_02_lb_0001.dds"},
{"3494405397.dds","pbw_02_ub_0001.dds"},
{"3521527687.dds","pew_00_lb_0045.dds"},
{"3527146215.dds","pbw_00_uw_0001_dec.dds"},
{"3533499275.dds","pew_00_ub_0045_01.dds"},
{"3534311588.dds","phw_00_sho_0031.dds"},
{"3545795712.dds","phw_00_lb_0031.dds"},
{"3573972132.dds","pvw_00_ub_0058_01.dds"},
{"3596470956.dds","pew_00_cloak_0032.dds"},
{"3609304591.dds","pkw_00_hand_0061.dds"},
{"3614655667.dds","phw_00_uw_0004_03_dec.dds"},
{"3630499470.dds","pew_00_ub_0003.dds"},
{"3654062849.dds","pvw_02_ub_0004_01.dds"},
{"365987242.dds","phw_01_sho_0002_dec.dds"},
{"374833862.dds","pew_00_hand_0010.dds"},
{"3780238344.dds","phw_02_sho_0002.dds"},
{"3789937337.dds","pbw_02_ub_0006_01.dds"},
{"3791029422.dds","pvw_00_ub_0058_02.dds"},
{"3862120745.dds","pwm_03_sho_0004.dds"},
{"3890606335.dds","pbw_00_ub_0046.dds"},
{"3919018533.dds","pew_00_uw_0030_02.dds"},
{"3941950328.dds","pew_10_sho_0010.dds"},
{"3950558239.dds","phw_00_lb_0013.dds"},
{"3958123004.dds","pew_00_ub_0034_01_hair.dds"},
{"397253889.dds","pew_00_lb_0008.dds"},
{"3988371908.dds","pvw_00_ub_0003.dds"},
{"4000169010.dds","phw_00_ub_0014.dds"},
{"401092744.dds","pbw_00_lb_0046_hair.dds"},
{"4058520570.dds","pew_00_ub_0038_dec.dds"},
{"4067792588.dds","phw_01_lb_0001.dds"},
{"4074048383.dds","pbw_00_lb_0015.dds"},
{"4086294970.dds","pvw_00_lb_0058_01.dds"},
{"4094405100.dds","pvw_00_foot_0000.dds"},
{"4105190757.dds","pvw_00_uw_0001_03_dec.dds"},
{"4117342010.dds","pbw_00_lb_0046_noshadow.dds"},
{"412326678.dds","phw_00_lb_0013_dec.dds"},
{"4131739567.dds","pew_00_sho_0030.dds"},
{"4134330014.dds","pbw_00_lb_0046.dds"},
{"4183875795.dds","pvw_00_sho_0001.dds"},
{"4195084077.dds","pew_00_uw_0031.dds"},
{"4197648293.dds","pbw_00_ub_0017.dds"},
{"4198233560.dds","pvw_00_ub_0052.dds"},
{"4260044339.dds","pew_00_lb_0032.dds"},
{"4261899337.dds","pvw_00_sho_0002.dds"},
{"4269153601.dds","pew_00_hand_0008.dds"},
{"4279038116.dds","pbw_02_lb_0002.dds"},
{"42915725.dds","pww_00_ub_0039.dds"},
{"491043312.dds","pkw_00_sho_0001_hair.dds"},
{"493059137.dds","pvw_02_lb_0002.dds"},
{"49710664.dds","pbw_00_lb_0018.dds"},
{"510760109.dds","pew_00_ub_0033.dds"},
{"532419852.dds","pew_00_ub_0031_hair.dds"},
{"550419656.dds","pew_00_ub_0010.dds"},
{"558783480.dds","pbw_00_hel_0046_hair.dds"},
{"569114108.dds","pew_00_uw_0030_dec.dds"},
{"596117228.dds","phw_00_ub_0031.dds"},
{"606975226.dds","pew_00_ub_0045.dds"},
{"637976496.dds","phw_00_ub_0007_dm.dds"},
{"657083173.dds","phw_01_sho_0003.dds"},
{"658433520.dds","pbw_02_lb_0003.dds"},
{"666619322.dds","phw_01_ub_0003.dds"},
{"666758142.dds","pew_00_hand_0003.dds"},
{"684067198.dds","pbw_03_ub_0002.dds"},
{"718885580.dds","pbw_03_lb_0002.dds"},
{"721980342.dds","phw_00_ub_0031_01.dds"},
{"766796120.dds","phw_02_ub_0005_02.dds"},
{"780345102.dds","pew_00_ub_0010_dm.dds"},
{"848997520.dds","pew_00_ub_0034.dds"},
{"868998973.dds","pvw_ew_hand_0009.dds"},
{"929321889.dds","phw_00_sho_0030.dds"},
{"94526970.dds","pvw_00_ub_0002.dds"},
{"958552097.dds","pvw_02_sho_0005.dds"},
{"963466424.dds","phw_00_uw_0006_03_dec.dds"},
{"972482231.dds","pew_00_lb_0033_dec.dds"},
{"992488401.dds","phw_00_ub_0039.dds"},
{"994933815.dds","pvw_00_sho_0003.dds"},
{"460973084.dds","m0025_impboss_ua_0002.dds"}
};

char* skipableFiles[] = {
"4067984630.dds",
"1022340771.dds",
"2618484627.dds",
"2592898066.dds",
"992151802.dds",
"1074882798.dds",
"985217646.dds",
"963036511.dds",
"882861563.dds",
"867726445.dds",
"831054362.dds",
"2766433969.dds",
"662197418.dds",
"2815313818.dds",
"570039615.dds",
"1248419673.dds",
"2528559386.dds",
"2518989139.dds",
"2902299366.dds",
"527715627.dds",
"1310308923.dds",
"2265589742.dds",
"2921113908.dds",
"1354249286.dds",
"1354295234.dds",
"1354405786.dds",
"3360555284.dds",
"1373017813.dds",
"1380358476.dds",
"2455257747.dds",
"1385013876.dds",
"2932259215.dds",
"473663678.dds",
"2452397449.dds",
"1398712904.dds",
"472539502.dds",
"460973084.dds",
"438306274.dds",
"2432761501.dds",
"2423145441.dds",
"1464921194.dds",
"4247759156.dds",
"1477890455.dds",
"4241677879.dds",
"4233968019.dds",
"4202232962.dds",
"1581019295.dds",
"4194355720.dds",
"161584728.dds",
"1642768885.dds",
"1643661565.dds",
"242032935.dds",
"1674467968.dds",
"4155760940.dds",
"4153684615.dds",
"3060502265.dds",
"309108375.dds",
"177398379.dds",
"1829412762.dds",
"1857053062.dds",
"2413765524.dds",
"4086396139.dds",
"3116457813.dds",
"1880007741.dds",
"2397860284.dds",
"3343119353.dds",
"3149741821.dds",
"3181258204.dds",
"1908889755.dds",
"1932263299.dds",
"3964613088.dds",
"3922727535.dds",
"1950091877.dds",
"3892679892.dds",
"376575145.dds",
"3765509177.dds",
"3761449123.dds",
"1982829551.dds",
"3736825072.dds",
"3727837277.dds",
"2023235948.dds",
"3719067796.dds",
"371496127.dds",
"3229738324.dds",
"3665314492.dds",
"2113744563.dds",
"3647750090.dds",
"3313692588.dds",
"2305135255.dds",
"3614657595.dds",
"3334807922.dds",
"3338776868.dds",
"2176226839.dds",
"2184122595.dds",
"3605462484.dds",
"34978033.dds",
"2276300242.dds"
};
