#include "../include/utilities.h"

int compare(void const *a, void const *b) {
    FileBlock *aa = (FileBlock *)a;
    FileBlock *bb = (FileBlock *)b;

    char* aaa = (char*)aa->fileName;
    char* bbb = (char*)bb->fileName;

    return strcmp(aaa, bbb);
}

int compare_meta_offset(void const *a, void const *b) {
    FileBlock *aa = (FileBlock *)a;
    FileBlock *bb = (FileBlock *)b;

    long aaa = (long)aa->metaOffset;
    long bbb = (long)bb->metaOffset;

    return (aaa - bbb);
}

int sort_by_folder_name_and_file_name(void const *a, void const *b) {
    FileBlock *aa = (FileBlock *)a;
    FileBlock *bb = (FileBlock *)b;

    char* folderNameA = (char*)aa->folderName;
    char* folderNameB = (char*)bb->folderName;

    char* fileNameA =  (char*)aa->fileName;
    char* fileNameB =  (char*)bb->fileName;

    char* completePathA = concatenate(folderNameA,fileNameA);
    char* completePathB = concatenate(folderNameB,fileNameB);

    int result = strcmp(completePathA, completePathB);

    free(completePathA);
    free(completePathB);

    return result;
}

long countPatchedFiles()
{
    printf("\nCounting patched files...");
    MetaFileInfo* metaFileInfo = getMetaFileInfo(getLatestBackup());
    long filesPatched = 0;
    int i = 0;
    FileBlock moddedFileBlocks;
    FILE* metaFile = openFile("pad00000.meta", "rb");

    fseek(metaFile,metaFileInfo->fileBlocksStart,SEEK_SET);

    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {
        fread(&moddedFileBlocks.hash,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.folderNum,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.fileNum,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.pazNum,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.fileOffset,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.zsize,sizeof(long),1,metaFile);
        fread(&moddedFileBlocks.size,sizeof(long),1,metaFile);

        if ((moddedFileBlocks.folderNum == RANDOM_FOLDER_NUM) && (moddedFileBlocks.fileNum == RANDOM_FILE_NUM))
        {
            filesPatched++;
        }
    }

    if (filesPatched == 0)
    {
        rewind(metaFile);
        long readNum1 = -1;
        long readNum2 = -1;
        fread(&readNum1,sizeof(long),1,metaFile);
        while(1)
        {
            if (!fread(&readNum2,sizeof(long),1,metaFile))
            {
                break;
            }
            if (readNum1 == RANDOM_FOLDER_NUM && readNum2 == RANDOM_FILE_NUM)
            {
                filesPatched++;
            }
            readNum1 = readNum2;
        }
    }
    fclose(metaFile);
    free(metaFileInfo);

    printf("\nDone.");

    return filesPatched;
}

char* endian_convert(int num)
{
    char big_end[12];
    sprintf(big_end,"%.8x", num);

    char* little_end = (char*) malloc(9);
    int i = 0;

    for(i = 0; i < strlen(big_end); i +=2)
    {
        char firstNum = big_end[strlen(big_end) - i - 2];
        if (firstNum >= 'a' && firstNum <= 'f')
        {
            firstNum -= 32;
        }
        char secondNum = big_end[strlen(big_end) - i - 1];
        if (secondNum >= 'a' && secondNum <= 'f')
        {
            secondNum -= 32;
        }
        little_end[i] = firstNum;
        little_end[i+1] = secondNum;
    }
    little_end[8]  = '\0';
    return little_end;
}


void printMainHeader()
{

    printf("******************************************************\n");
    printf("*                                                    *\n");
    printf("*                Meta Injector v2.1                  *\n");
    printf("*                                                    *\n");
    printf("******************************************************\n");
}


char* getTwoDigits(char* fileName)
{
    int start = indexOf('_',fileName,0) + 1;
    int end = indexOf('_',fileName,1);

    char* twoDigits = substr(fileName,start,end);

    return twoDigits;
}

char* getMiddleDigits(char* fileName)
{
    int  start = indexOf('_',fileName,1) + 1;
    int end = indexOf('_',fileName,2);

    char* middle = substr(fileName,start,end);

    return middle;
}

char* getClassPrefix(char* fileName)
{
    return substr(fileName,0,3);
}

char* getLastFolder(char* folderName)
{
    int i = strlen(folderName);
    int start = 0;
    int skipped = 0;
    for ( i = strlen(folderName); i >=0; i--)
    {
        if (folderName[i] == '/')
        {
            if (skipped == 0)
            {
                skipped = 1;
            }
            else
            {
                start = i + 1;
                break;
            }
        }
    }
    return substr(folderName,start,strlen(folderName) - 1);
}

int contains(char* fullString, char* substring)
{
    if (strlen(substring) > strlen(fullString))
    {
        //printf("\nWarning: Wrong use of function 'contains(%s,%s)'. The first argument has to be longer than the second", fullString, substring);
        return 0;
    }
    int i = 0;
    int j = 0;

    int combo = 0;
    int starCount = 0;
    for (i = 0; i < strlen(fullString); i++)
    {
        if (substring[j] == '*')
        {
            //Ignores the next character in the "fullstring"
            i++; // This prevents things like this being true: _003 contains _*003

            //printf("\n\nsubstring[%d] == \'*\'",j);
            //printf("\nwhile substring[%d] == \'*\'",j);
            while(substring[j] == '*')
            {
                starCount++;
                //printf("\nstarCount = %d", starCount);
                j++;
               //printf("\nj = %d", j);
                if (j >= strlen(substring))
                {
                   // printf("\nj >= strlen(substring)");
                    //printf("\nbreak;");
                    if (combo == (strlen(substring) - starCount))
                    {
                        //printf("\n\n RETURN 1");
                       // PAUSE();
                        return 1;
                    }
                    else
                    {
                        return 0;
                    }
                }
            }
           //printf("\nout while1.\n\nfullString[i] = %c     substring[j] = %c", fullString[i],substring[j]);

            if (j < strlen(substring))
            {
              // printf("\nwhile fullString[i] = %c  != substring[j] = %c", fullString[i],substring[j]);
                while(fullString[i] != substring[j])
                {
                    //printf("\nfullString[i] = %c  substring[j] = %c", fullString[i],substring[j]);
                    i++;
                   //printf("\ni = %d", i);
                    if (i >= strlen(fullString))
                    {
                        //printf("\nj >= strlen(substring)");
                       //printf("\nbreak;");
                        break;
                    }
                }

               //printf("\noutwhile2\n\n");
            }
        }
        if (substring[j] == fullString[i]
        || ((substring[j] + 32  <= 'z') && (substring[j] + 32 == fullString[i]))
        || ((substring[j] - 32 >= 'A') && (substring[j] - 32 == fullString[i]))
        || (substring[j] == '/' && fullString[i] == '\\')
        || (substring[j] == '\\' && fullString[i] == '/'))
        {
           //printf("\n\nNormal: fullstring[i] = %c  == %c substring[j]   i: %d  j : %d", fullString[i],substring[j],i,j);
            combo++;
            if (j + 1 <= strlen(substring))
            {
                j++;
            }
            else
            {
                combo = 0;
                j = 0;
                starCount = 0;
               //printf("\n\nj(%d) + 1 <= strlen(substring)(%d)", j,strlen(substring));
               //printf("\ncombo break 2   combo = 0 j = 0\n\n");
            }
        }

        else
        {
            //printf("\n\nNormal: fullstring[i] = %c  != %c substring[j]   i: %d  j : %d", fullString[i],substring[j],i,j);
            combo = 0;
            j = 0;
            starCount = 0;
           //printf("\ncombo break 1  combo = 0 j = 0\n\n");
        }
        //printf("\ncheck combo == strlen(substring) - starcount = %d  == %d - %d  (%d)  ", combo,strlen(substring),starCount,strlen(substring) - starCount);
        if (combo == (strlen(substring) - starCount))
        {
           //printf("\n\n RETURN 1");
            //PAUSE();
            return 1;
        }
    }
    //printf("\n\n RETURN 0");
    //PAUSE();
    return 0;
}


char* getFourDigits(char* fileName)
{
    int start = indexOf('_',fileName,2) + 1;
    int end = indexOf('_',fileName,3);
    int dot = indexOf('.',fileName,0);

    if (dot < end)
    {
        end = dot;
    }

    char* fourDigits = substr(fileName,start,end);

    return fourDigits;
}


void askConfirmation()
{
    PAUSE();
    /*printf("Do you want to proceed anyway? (y/n) ");
    char input[32] = "";
     while (strcmpi(input,"y") != 0 && strcmpi(input,"n") != 0)
     {
        fflush(stdin);
        scanf("%s", input);
        fflush(stdin);
     }

    if (strcmpi(input,"n") == 0)
    {
        printf("\nExiting.\n\n");
        system("PAUSE");
        exit(1);
    }*/
}

void createPath(char* pathToCreate)
{
    charReplace(pathToCreate,'/','\\');

    DIR* dir = opendir(pathToCreate);

    if (dir)
    {
        closedir(dir);
    }
    else
    {
        char* mdCommand = (char*) malloc(strlen("md \"%s\"") + strlen(pathToCreate) + 1);
        sprintf(mdCommand,"md \"%s\"",pathToCreate);
        //printf("\n%s\n",mdCommand);
        system(mdCommand);
        free (mdCommand);
    }
}

void backupMenu()
{
    int backupSelected = -1;

        long backupCount = 0;
        char** backupNames = getBackupList(&backupCount);
        if (backupCount > 0)
        {
            while (1)
            {
                 backupSelected  = selectBackup(backupNames, backupCount);

                 if (backupSelected == -1)
                 {
                     return;
                 }

                 if (backupSelected < 0 || backupSelected > backupCount - 1)
                 {
                    printf("\nInvalid option!\n\n");
                 }
                 else
                 {
                     break;
                 }
            }

            restoreBackup(backupNames[backupSelected]);
        }
        else
        {
             printf("No backup was found");
        }
        PAUSE();
}



char* getLatestBackup()
{
    long backupCount = 0;
    char** backups = getBackupList(&backupCount);
    if (backupCount == 0)
    {
        return "pad00000.meta";
    }
    return backups[backupCount - 1];
}

// Merges 2 strings into 1
char* concatenate (char* str1, char*str2)
{
    char* result = (char*)malloc(strlen(str1) + strlen(str2) + 1);
    strcpy(result,str1);
    strcat(result,str2);
    return result;
}


int isNumOrLetter(char c)
{
    if (c >= 'a' && c <='z')
    {
        return 1;
    }
    if (c >= 'A' && c <='Z')
    {
        return 1;
    }
    if (c >= '0' && c <='9')
    {
        return 1;
    }
    if (c == '_')
    {
        return 1;
    }
    return 0;
}

int hexToInt(char* hex)
{
   /*FILE* tmpFile = fopen("tmp.bin","wb");
   fwrite(hex,sizeof(int),1,tmpFile);
   int num = 0;
   rewind(tmpFile);
   fread(tmpFile)*/
   return 0;
}

char* intToHex(int x)
{
    char* hex = (char*) malloc(sizeof(int) + 1);
    sprintf(hex,"%d",x);
    return hex;
}

void printColor(char* stringToPrint, Color COLOR)
{
    if (isWindows10)
    {
        switch (COLOR)
        {
            case (RED):
            {
                printf("\033[31;1m%s\033[0m", stringToPrint);
            }break;
            case (GREEN):
            {
                printf("\033[32;1m%s\033[0m", stringToPrint);
            }break;
            case (YELLOW):
            {
                printf("\033[33;1m%s\033[0m", stringToPrint);
            }break;
        }
    }
    else
    {
        printf("%s", stringToPrint);
    }
}

int getWindowsVersion()
{
    createPath("patcher_resources\\");
    system("cd patcher_resources\\ & ver >windows_version.txt");

    FILE* fp = fopen("patcher_resources\\windows_version.txt", "rb");

    if (fp == NULL)
    {
        return 0;
    }
    int c = '\0';
    while(fread(&c,1,1,fp))
    {
        if (c == '.')
        {
            fseek(fp,-3,SEEK_CUR);
            char versionStr[3] = "";
            fread(versionStr,2,sizeof(char),fp);
            fclose(fp);
            remove("patcher_resources\\windows_version.txt");
            return  atoi(versionStr);

        }
    }
    fclose(fp);
    remove("patcher_resources\\windows_version.txt");
    return 0;
}

void listPatchedFiles(int patchedFilesCount)
{
    if (patchedFilesCount == 0)
    {
        printf("\n  No files patched.");
        PAUSE();
        return;
    }

    int i = 0;

    MetaFileInfo* metaFileInfo = getMetaFileInfo(getLatestBackup());
    FileBlock* fileBlocks = fillFileBlocks(metaFileInfo);

    FileBlock* patchedFiles = (FileBlock*) malloc(patchedFilesCount * sizeof(FileBlock));


    int currentPatchedFile = 0;



    FILE* metaFile = openFile("pad00000.meta", "rb");
    printf("Searching patched files...");
    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {
        if (isPatched(&fileBlocks[i],metaFileInfo,metaFile))
        {
            memcpy (&patchedFiles[currentPatchedFile], &fileBlocks[i], sizeof(FileBlock));
            currentPatchedFile++;
        }
    }
    printf("\nDone.");
    fclose(metaFile);
    printf("\n\nSorting patched files...");
    qsort(patchedFiles,patchedFilesCount,sizeof(FileBlock),sort_by_folder_name_and_file_name);
    printf("\nDone.");
    system("cls");
    printMainHeader();
    printf("List of currently patched files:\n");
    for(i = 0; i < patchedFilesCount; i++)
    {
        if (patchedFilesCount < 10)
        {
            printf("%d", i+1);
        }
        else if (patchedFilesCount < 100)
        {
            printf("%.2d", i+1);
        }
        else if (patchedFilesCount < 1000)
        {
            printf("%.3d", i+1);
        }
        else
        {
            printf("%d", i+1);
        }
        printf(" - %s%s\n", patchedFiles[i].folderName, patchedFiles[i].fileName);
    }

    PAUSE();
}


void printFileBlock(FileBlock fileBlock)
{
    printf("\n%ld|%ld|%ld|offset:%ld|%s|%s",fileBlock.hash,fileBlock.folderNum,fileBlock.fileNum,fileBlock.metaOffset,fileBlock.folderName,fileBlock.fileName);
}

void addToStringArray(char* fileNameToAdd, char*** ref_stringArray, int* ref_arrraySize)
{
    // If the variable that holds the fileNames of the files that failed to patch hasn't been initialized yet
    if ((*ref_stringArray) == NULL)
    {
        (*ref_stringArray) = (char**) malloc(sizeof(char*));
    }
    else // If it has been initialized
    {
         // Increases the size
        (*ref_stringArray) = (char**) realloc((*ref_stringArray), (((*ref_arrraySize) + 1) * sizeof(char*)));
    }
    // Allocates memory to copy the failed file name
    (*ref_stringArray)[(*ref_arrraySize)] = (char*) malloc(strlen(fileNameToAdd) + 1);
    strcpy((*ref_stringArray)[(*ref_arrraySize)],fileNameToAdd); // Copies the name of the failed file

    (*ref_arrraySize)++;
}

int isPatched(FileBlock* fileBlock, MetaFileInfo* metaFileInfo, FILE* metaFile)
{
    fseek(metaFile,fileBlock->metaOffset,SEEK_SET);
    int hashFound = 0;
    long hash = -1;
    fread(&hash,sizeof(long),1,metaFile);
    if (hash != fileBlock->hash)
    {
        while(ftell(metaFile) >= metaFileInfo->fileBlocksStart)
        {
            fseek(metaFile,((-2) * sizeof(long)), SEEK_CUR);
            fread(&hash,sizeof(long),1,metaFile);
            if (hash == fileBlock->hash)
            {
                hashFound = 1;
                break;
            }
        }
        if (hashFound == 0)
        {
            fseek(metaFile,fileBlock->metaOffset + sizeof(long),SEEK_SET);
            while(ftell(metaFile) < metaFileInfo->fileBlocksEnd)
            {
                fread(&hash,sizeof(long),1,metaFile);
                if (hash == fileBlock->hash)
                {
                    hashFound = 1;
                    break;
                }
            }
        }
    }
    else
    {
        hashFound = 1;
    }
    if (hashFound == 0)
    {
        return 0;
    }
    else
    {
        long folderNum = -1;
        long fileNum = -1;
        fread(&folderNum,sizeof(long),1,metaFile);
        fread(&fileNum,sizeof(long),1,metaFile);

        if (folderNum == RANDOM_FOLDER_NUM && fileNum == RANDOM_FILE_NUM)
        {
            return 1;
        }
        else
        {
            return 0;
        }
    }
    return 0;
}

int alreadyInStringArray(char* fileName,char** stringArray, int arrraySize)
{
    int i = 0;
    for (i = 0; i < arrraySize; i++)
    {
        if (strcmpi(stringArray[i],fileName) == 0)
        {
            return 1;
        }
    }
    return 0;
}

void printHex(char* str)
{
    int i = 0;
    printf("\n");
    for (i = 0; i < strlen(str); i++)
    {
        printf("%x ", str[i]);
    }
    printf("\n");
}


void charReplace(char* str,char token,char replace)
{
    int i = 0;
    for (i = 0; i < strlen(str); i++)
    {
        if (str[i] == token)
        {
            str[i] = replace;

        }
    }
}

void printNextAsHex(char* str,int start, int length)
{
    int i = 0;

    for (i = start; i < (start + length); i++)
    {
        printf("%x ", str[i]);
    }
    printf("\n");
}

void printNextAsDec(char* str,int start, int length)
{
    int i = 0;
    printf("\n");
    for (i = start; i < (start + length); i++)
    {
        printf("%d ", str[i]);
    }
    printf("\n");
}

void printIntAsHex(int num)
{
    char* converted = intToHex(num);
    printHex(converted);
}

int nextFourBytesAsInt(char* str, int startLocation)
{
    int i = 0;
    char fourBytes[5] = "";
    fourBytes[4] = '\0';
    for (i = 0; i < 4; i++)
    {
        fourBytes[i] = str[startLocation+i];
    }
    return hexToInt(fourBytes);
}

char* substr(char* str,int start, int end)
{
    if (end < start)
    {
        //printf("\ntrying to create a substring where end < start |  %s start: %d   end: %d\n", str, start, end);
        return "";
    }
    if (end == start)
    {
        return "";
    }
    char* sub = (char*) malloc((end - start) + 1);
    int j = 0;
    int i = 0;
    for (i = start; i <= end; i++)
    {
        sub[j] = str[i];
        j++;
    }
    sub[j - 1] = '\0';
    return sub;
}

int indexOf(char token, char* str, int skips)
{
    int i = 0;
    int skipped = 0;
    for (i = 0; i < strlen(str); i++)
    {
        if (str[i] == token)
        {
            if (skipped == skips)
            {
                return i;
            }
            else
            {
                skipped++;
            }
        }
    }
    return i -1;
}

long locateHash(long hash,MemFileInt* metaFileInMemory,MetaFileInfo* metaFileInfo)
{
    long currentBlock = 0;
    int readNum = 0;

    int startBlock = metaFileInfo->fileBlocksStart/sizeof(int);
    int endBlock = metaFileInfo->fileBlocksEnd/sizeof(int);

    for (currentBlock = startBlock; currentBlock < endBlock; currentBlock++)
    {
        readNum = metaFileInMemory->block[currentBlock];

        if(readNum == hash)
        {
            return currentBlock;
        }
    }
    return NOT_FOUND;
}


long locateIntInMemory(int intToLocate,MemFileInt* arrayInMemory, MetaFileInfo* metaFileInfo)
{
    long currentBlock = 0;
    int readNum = 0;

    for (currentBlock = 0; currentBlock < arrayInMemory->size; currentBlock++)
    {
        readNum = arrayInMemory->block[currentBlock];


        int startBlock = metaFileInfo->fileBlocksStart/sizeof(int);
        int endBlock = metaFileInfo->fileBlocksEnd/sizeof(int);


        if(readNum == intToLocate)
        {
            if ((currentBlock < startBlock || currentBlock > endBlock))
            {
                printf("                       OUTSIDE!");
                 return NOT_FOUND;
            }
            return currentBlock;

        }

    }
    return NOT_FOUND;
}

long locate2IntsInMemory(int int1, int int2 ,MemFileInt* arrayInMemory)
{
    //printf("\nlooking for: %.8x | %.8x\n", int1, int2);
    long currentBlock = 0;
    for (currentBlock = 0; currentBlock < arrayInMemory->size; currentBlock++)
    {

        if(arrayInMemory->block[currentBlock] == int1 && arrayInMemory->block[currentBlock + 1] == int2)
        {
            return currentBlock;
        }

    }
    return NOT_FOUND;
}

void undoLastChanges()
{

    FILE* metaFile = openFile("pad00000.meta","rb+");
    FILE* modFile = fopen("patcher_resources\\latest_modifications_meta_injector.bin","rb");
    if (modFile == NULL)
    {
        printf("\nNo last modifications detected.\n\n");
        fclose(metaFile);
        system("PAUSE");
        return;
    }

    printf("\n\nReverting File");
    if (getFileSize(modFile) == 0)
    {
        printf("\nNo changes to undo.\n\n");
        fclose(metaFile);
        fclose(modFile);
        system("PAUSE");
        return;
    }
    else
    {
        MetaFileInfo* metaFileInfo = getMetaFileInfo("pad00000.meta");
        fseek(metaFile,metaFileInfo->fileBlocksStart,SEEK_SET);;

        long offset = 0;
        long folderNum = 0;
        long fileNum = 0;
        long changesCount = 0;


        while(fread(&offset,sizeof(long),1,modFile))
        {
            fread(&folderNum,sizeof(long),1,modFile);
            fread(&fileNum,sizeof(long),1,modFile);

            fseek(metaFile,offset + sizeof(long),SEEK_SET);

            fwrite(&folderNum,sizeof(long),1,metaFile);
            fwrite(&fileNum,sizeof(long),1,metaFile);

            changesCount++;
        }

        printf("\nReverted changes: %ld\n\n", changesCount);



        fclose(modFile);
          remove("patcher_resources/latest_modifications_meta_injector.bin");
        fclose(metaFile);
        free(metaFileInfo);
        system("PAUSE");
    }
}


/*void toHex(int decimalNumber)
{
	int remainder;
	int quotient;
    int i=1,j;
    char hexadecimalNumber[100];

    quotient = decimalNumber;

    while(quotient!=0)
	{
         remainder = quotient % 16;

      //To convert integer into character
      if( remainder < 10)
           remainder = remainder + 48;		// Add 48(become ascii later) value to remainder if less than 10--see ascii table for more info
      else
         remainder = remainder + 55;		// Add 55(become ascii later) value to remainder if greater than 10--see ascii table for more info

      hexadecimalNumber[i++]= remainder;

      quotient = quotient / 16;
    }

    //printf("Equivalent hexadecimal value of %d is: ",decimalNumber);

    for(j = i -1 ;j> 0;j--)
    {
      printf("%c",hexadecimalNumber[j]);
    }

}

void printNextBytesAsHex(FILE* fp, int howManyBytes)
{
    int i = 0;
    char c;
    for (i = 0; i < howManyBytes; i++)
    {
        fread(&c,1,1,fp);
        printf("%x ", c);
    }
}*/



char** getBackupList(long* backupCount)
{
    return getFilesSingleFolder(getCurrentPath(),"backup",backupCount);
}

int selectBackup(char** backupNames, long backupCount)
{
    long i = 0;
     char input[32];
     int optionSelected = -1;
    int backupSelected = -1;

    printf("\n\nList of backups:\n");
    for (i = 0; i < backupCount; i++)
    {
       printf("%ld - %s\n", i + 1, backupNames[i]);
    }
    printf("0 - Back\n");
    while (1)
    {
        printf("\nSelect backup number: ");
        fgets(input,32,stdin);

        optionSelected = atoi(input);
        backupSelected = optionSelected - 1;

        if (optionSelected == EXIT)
        {
            return -1;
        }

        if (optionSelected < 0 || optionSelected > backupCount)
        {
            printf("\nInvalid option!\n\n");
        }
        else
        {
            break;
        }
    }

    return backupSelected;
}



void createBackup()
{
    int backupNameLen = strlen("pad00000[YYYY-MM-DD].meta.backup");
    char* backupName = (char*)malloc(backupNameLen + 1);


    remove("patcher_resources/latest_modifications_meta_injector.bin");

    time_t now;
    time(&now);

    struct tm* now_tm;
    now_tm = localtime(&now);

    strftime (backupName, backupNameLen + 1, "pad00000[%Y-%m-%d].meta.backup", now_tm);

    char* cmd = (char*) malloc(strlen("copy pad00000.meta pad00000[YYYY-MM-DD].meta.backup") + 1);
    sprintf(cmd,"copy pad00000.meta %s", backupName);
    system(cmd);

    if (fileExists(backupName))
    {
        printf("\nA new backup was created: %s\n",backupName);
    }
    else
    {
        printf("\nCould not create a backup\n");
        printf("Try running the program as administrator\n");
        printf("\nThe program will try to ignore this file, but it will most likely, not work properly.");
        PAUSE();
    }
}

int backupExists()
{
    long backupCount = 0;
    char** backups = getFilesSingleFolder(getCurrentPath(),"backup", &backupCount);

    if (backupCount == 1)
    {
        printf("%ld backup was found\n", backupCount);
    }
    else
    {
        printf("%ld backups were found\n", backupCount);
    }

    FILE* backupFile = NULL;
    FILE* metaFile = openFile("pad00000.meta","rb");

    int i = 0;
    for(i = 0; i < backupCount; i++)
    {
        backupFile = fopen(backups[i],"rb");
        if (getFileSize(backupFile) == getFileSize(metaFile))
        {
            fclose(metaFile);
            fclose(backupFile);
            return 1;
        }
        fclose(backupFile);
    }
    fclose(metaFile);
    return 0;
}

void restoreBackup(char* backupName)
{
    char command[2048];
    sprintf(command,"xcopy /y \"%s\" pad00000.meta",backupName);
    system(command);
    preventFileRecheck();
    remove("patcher_resources\\latest_modifications_meta_injector.bin");

}

//-----------------------------------------------------------------------------------------------------------------
void createLogFile()
{
    FILE* log = openFile("patcher_resources/paz_browser.log","wb");
    long lastMetafileSize = getFileSizeByName("pad00000.meta");
    fwrite(&lastMetafileSize,sizeof(long),1,log);
    fclose(log);
}

char* bdoRootFolder()
{
    return (oneLevelDown(getCurrentPath()));
}

int metaFileChangedSize()
{
    long lastMetafileSize = 0;
    FILE* log = fopen("patcher_resources/paz_browser.log","rb");
    if (!log)
    {
        return 1;
    }
    fread(&lastMetafileSize,sizeof(long),1,log);

    if (lastMetafileSize != getFileSizeByName("pad00000.meta"))
    {
        fclose(log);
        return 1;
    }
    fclose(log);
    return 0;
}


int createMenu(MenuElements menu)
{
    char input[2048] = "";       // A simple buffer to store whatever the user types in the menu
    int optionSelected = -1;
    int i = 0;

     // Display the menu to select the region
    while(1) // This condition makes the menu repeat itself until a valid input is entered
    {
        system("cls"); // clears the screen
        printMainHeader();
        printf("%s\n", menu.header);
        for (i = 1; i <= menu.nOptions; i++)
        {
            printf("%d - %s\n",i, menu.menuoptions[i]);
        }
        printf("0 - %s\n", menu.menuoptions[0]);
        printf("\nEnter your choice: ");
        fflush(stdin);
        fgets(input,2048,stdin);
        input[strlen(input) - 1] = '\0';

        // Converts the read string to int
        optionSelected = atoi(input);

        // This variable also tells the program later which one of the "moddedBytes" the program should use when replacing the bytes

        if (optionSelected == 0)
        {
            return 0;
        }

         // If the user entered an invalid option, displays a error message
        if(optionSelected < 1 || optionSelected > menu.nOptions)
        {
            printf("\nInvalid option!\n\n");
            Sleep(500);
        }
        else
        {
            break;
        }
    }

    return optionSelected;
}

FileBlock* binarySearchFileBlock(char* key, FileBlock* fileBlocks, int fileblocksCount)
{
    int first = 0;
    int last = fileblocksCount - 1;
    int middle = (int)(first+last)/2;

    while (first <= last)
    {
        if (strcmpi(fileBlocks[middle].fileName,key) < 0)
        {
            first = middle + 1;
        }

        else if (strcmpi(fileBlocks[middle].fileName,key)  == 0)
        {
            return &fileBlocks[middle];
        }
        else
        {
            last = middle - 1;
        }

        middle = (first + last)/2;
   }
   if (first > last)
   {
      /*printf("Not found! %s is not present in the list.\n", key);
      system("PAUSE");*/
   }
   return NULL;
}

void PAUSE()
{
    printf("\n\n");
    system("PAUSE");
}

int createArrowMenu(MenuElements menu, char** subtittle, int defaultSelected)
{
    char input = '\0';
    int cursor = 0;
    if (defaultSelected < menu.nOptions && defaultSelected >= 0)
    {
        cursor = defaultSelected;
    }
    int i = 0;

    while(1)
    {
        system("cls");
        printMainHeader();
        printf("%s\n\n", menu.header);
        for (i = 0; i < menu.nOptions; i++)
        {
            if(cursor == i)
            {
               printf(" -> %s\n",menu.menuoptions[i]);
            }
            else
            {
                printf("    %s\n",menu.menuoptions[i]);
            }
        }

           printf("\n%s\n", subtittle[cursor]);

        if (strcmpi(menu.menuoptions[i-1],"Exit") != 0)
        {
            printf("\nENTER     - Confirm");
            printf("\nARROWS    - Change Selected Option");
            printf("\nBACKSPACE - Go back\n");
        }
        else
        {
            printf("\n");
            printf("ENTER  - Confirm\n");
            printf("ARROWS - Change Selected Option\n");
            printf("ESC    - Exit\n");
        }

        input = getch();

        if (input == ENTER)
        {
            break;
        }
        else if (input == ARROW_DOWN)
        {
            if (cursor + 1 < menu.nOptions)
            {
                cursor++;
            }
            else
            {
                cursor = 0;
            }
        }
        else if (input == ARROW_UP)
        {
            if (cursor - 1 >= 0)
            {
                cursor--;
            }
            else
            {
                cursor = menu.nOptions;
            }
        }
        if (strcmpi(menu.menuoptions[i-1],"Exit") != 0)
        {
            if (input == BACKSPACE)
            {
                return BACKSPACE;
            }
        }
        else
        {
            if (input == ESC)
            {
                return EXIT;
            }
        }
    }

    return cursor;
}


void preventFileRecheck()
{
    printf("\nPreventing File Check...\n");
    char* root = bdoRootFolder();

    char* configFileCheckPath = concatenate(root,"config.filecheck");
    char* versionDatPath = concatenate(root,"version.dat");


    FILE* metaFile = openFile(getLatestBackup(),"rb");
    FILE* versionDatFile = fopen(versionDatPath,"rb");


    int clientVersion_versionDat = 0;
    int clientVersion_metaFile = 0;



    fread(&clientVersion_metaFile,sizeof(int),1,metaFile);
    if (versionDatFile != NULL)
    {
         long versionDatFileSize = getFileSize(versionDatFile);
         char* versionDatAsString = (char*) malloc(versionDatFileSize);

        fread(versionDatAsString,sizeof(char),versionDatFileSize,versionDatFile);
        fclose(versionDatFile);

        versionDatAsString[versionDatFileSize-1] = '\0';

        clientVersion_versionDat = atoi(versionDatAsString);
        free(versionDatAsString);

        if (clientVersion_versionDat != clientVersion_metaFile)
        {
            char proceed = 0;
            printf("\nWarning: Client version from version.dat (%d) differs from the pad00000.meta file (%d).", clientVersion_versionDat,clientVersion_metaFile);
            printf("\nAre you sure you want to proceed? (y/n)\n");
            printf(":");
            scanf("%c", &proceed);

            if (proceed != 'y' && proceed != 'Y')
            {
                free(configFileCheckPath);
                free(versionDatPath);
                free(root);
                fclose(metaFile);
                printf("\nNo.");
                return;
            }
            printf("\nYes.");
        }
    }

    if (!remove(configFileCheckPath))
    {
        printf("\nconfig.filecheck - successfully deleted.\n");
    }

    printf("\nCreating version.dat file...\n");
    versionDatFile = openFile(versionDatPath,"wb");
    if (versionDatFile != NULL)
    {
        char endByte = '\x0A';
        char clientVersionAsString[16];
        sprintf(clientVersionAsString,"%d",clientVersion_metaFile);
        fwrite(&clientVersionAsString,sizeof(char),strlen(clientVersionAsString),versionDatFile);
        fwrite(&endByte,1,1,versionDatFile);
        fclose(versionDatFile);
    }
    printf("Done.");
    free(configFileCheckPath);
    free(versionDatPath);
    free(root);
    fclose(metaFile);

}

void openFolder(char* folderToOpen)
{
    charReplace(folderToOpen,' ','_');
    char* command = (char*) malloc(strlen("start \"\"") + strlen(folderToOpen) + 1);
    sprintf(command,"start %s",folderToOpen);
    printf("\n%s",command);
    system(command);
    free(command);
}


int beginsWith(char* fullString, char* substring)
{
    if (strlen(substring) > strlen(fullString))
    {
        //printf("\nWarning: Wrong use of function 'contains(%s,%s)'. The first argument has to be longer than the second", fullString, substring);
        return 0;
    }
    int i = 0;
    int j = 0;
    int combo = 0;
    for (i = 0; i < strlen(fullString); i++)
    {
        if (substring[j] == fullString[i]
        || ((substring[j] + 32  <= 'z') && (substring[j] + 32 == fullString[i]))
        || ((substring[j] - 32 >= 'A') && (substring[j] - 32 == fullString[i]))
        || (substring[j] == '/' && fullString[i] == '\\')
        || (substring[j] == '\\' && fullString[i] == '/'))
        {
            combo++;
            if (j + 1 <= strlen(substring))
            {
                j++;
            }
            else
            {
                return 0;
                j = 0;
            }

        }
        else
        {
            return 0;
            j = 0;
        }
        if (combo == strlen(substring))
        {
            return 1;
        }
    }
    return 0;
}

int extractFileCustomPath(FileBlock* fileBlockSelected, char* extractionLocation, int echo)
{
    charReplace(extractionLocation,' ','_');
    if(echo)
    {
        printf("\n\nExtracting %s to: %s\n", fileBlockSelected->fileName,extractionLocation);
        printf("\n[FILE INFO]  :");
        printf("\nFile Num   : %ld", fileBlockSelected->fileNum);
       /* if (fileBlockSelected->hasRealName)
        {
            printf("\nReal Name  : %s", fileBlockSelected->RealName);
        }*/
        printf("\nFile Name  : %s", fileBlockSelected->fileName);
        printf("\nFolder Name: %s", fileBlockSelected->folderName);
        printf("\nPAZ: PAD%.5ld.PAZ\n\n", fileBlockSelected->pazNum);



    }

     if (!fileExists("patcher_resources\\quickbms.exe"))
    {
        printf("\n\nquickbms.exe is not located in the \"patcher_resources\\\" folder\n\n");
        system("PAUSE");
        free(extractionLocation);
        return 0;
    }

    char pazName[32];
    sprintf(pazName,"PAD%.5ld.PAZ",fileBlockSelected->pazNum);
    char* command = (char*) malloc(strlen("cd patcher_resources\\ & quickbms.exe  1>NULL 2>NULL2 -Y -f \"*%s*\" simple_extraction.bms \"../%s\" \"../%s") + strlen(fileBlockSelected->fileName) + strlen(pazName) + strlen(extractionLocation) + 256);
    if (echo)
    {
        sprintf(command,"cd patcher_resources\\ & quickbms.exe -Y -f \"*%s*\" simple_extraction.bms \"../%s\" \"../%s", fileBlockSelected->fileName,pazName,extractionLocation);
        //printf("command: %s\n", command);
    }
    else
    {
        sprintf(command,"cd patcher_resources\\ & quickbms.exe  1>NULL 2>NULL2 -Y -f \"*%s*\" simple_extraction.bms \"../%s\" \"../%s", fileBlockSelected->fileName,pazName,extractionLocation);
    }

    system(command);

    char* destPath = concatenate(extractionLocation,fileBlockSelected->fileName);

    //charReplace(destPath,'/','\\');

    if (!fileExists(destPath))
    {
        printf("\nExtracting \"%s\" to: \"%s\"\n", fileBlockSelected->fileName,extractionLocation);
        printf("\nError: Could not extract file.\n");
        sprintf(command,"cd patcher_resources\\ & quickbms.exe  1>NULL 2>NULL2 -Y -f \"*%s*\" simple_extraction.bms \"../%s\" \"../%s", fileBlockSelected->fileName,pazName,extractionLocation);
        system(command);
        free(command);
        FILE* quickbmsOutputFile = fopen("patcher_resources\\NULL2", "rb");

        if (quickbmsOutputFile != NULL)
        {
            char* quickbmsOutputFileInMemory = (char*)malloc(getFileSize(quickbmsOutputFile) + 1);
            fread(quickbmsOutputFileInMemory,getFileSize(quickbmsOutputFile),sizeof(char),quickbmsOutputFile);
            fclose(quickbmsOutputFile);

            if (contains(quickbmsOutputFileInMemory,"fdnum_open()"))
            {
                printf("Reason: Could not find or open the file %s in your \"Paz\" folder\n\n", pazName);
            }
        }
        remove("patcher_resources\\NULL");
        remove("patcher_resources\\NULL2");
        system("PAUSE");
        return 0;
    }
    if (echo)
    {
        printf("\nSuccess: File extracted: \"%s\"\n",destPath);
    }

    free(command);
    remove("patcher_resources\\NULL");
    remove("patcher_resources\\NULL2");
    return 1;
}

int containsStartEnd(char* fullString, char* substring, int* start, int* end)
{
    if (strlen(substring) > strlen(fullString))
    {
        //printf("\nWarning: Wrong use of function 'contains(%s,%s)'. The first argument has to be longer than the second", fullString, substring);
        return 0;
    }
    int i = 0;
    int j = 0;
    int combo = 0;
    for (i = 0; i < strlen(fullString); i++)
    {
        if (substring[j] == fullString[i]
        || substring[j] + 32 == fullString[i]
        || substring[j] - 32 == fullString[i]
        || (substring[j] == '/' && fullString[i] == '\\')
        || (substring[j] == '\\' && fullString[i] == '/'))
        {
            combo++;
            if (j + 1 <= strlen(substring))
            {
                j++;
            }
            else
            {
                combo = 0;

                j = 0;
            }

        }
        else
        {
             (*start) = i + 1;
            combo = 0;
            j = 0;
        }
        if (combo == strlen(substring))
        {
            (*end) = (*start) + strlen(substring);
            return 1;
        }
    }
    return 0;
}


